/*
 * ffmpeg2theora.c -- Convert ffmpeg support a/v files to theora
 * Version 0.6
 * Copyright (C) 2003-2004 <j@thing.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 */
#include "common.h"
#include "avformat.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include "theora/theora.h"
#include "vorbis/codec.h"
#include "vorbis/vorbisenc.h"

#include "theorautils.h"

FILE *outfile;
theoraframes_info info;
static int using_stdin = 0;

typedef struct ff2theora{
	AVFormatContext *context;
	int video_index;
	int audio_index;
	int frequency;
	int channels;
	int scale;
	int wide;
	int twopass;
	int preview;
	double start;
	double end;
	int audio_output;
	char *filter;
	int scaler;
	int output_width;
	int output_height;
	int every;
	int disable_audio;
	double fps;
	int audio_resample;
	ImgReSampleContext *img_resample_ctx; /* for image resampling */
	/* some more config options */
	int video_quality;
	int audio_quality;
    
}
 *ff2theora;

/** Allocate and initialise an AVFrame.
*/
AVFrame *alloc_picture (int pix_fmt, int width, int height){
	AVFrame *picture;
	uint8_t *picture_buf;
	int size;

	picture = avcodec_alloc_frame ();
	if (!picture)
		return NULL;
	size = avpicture_get_size (pix_fmt, width, height);
	picture_buf = av_malloc (size);
	if (!picture_buf){
		av_free (picture);
		return NULL;
	}
	avpicture_fill ((AVPicture *) picture, picture_buf,
			pix_fmt, width, height);
	return picture;
}

ff2theora ff2theora_init (){
	ff2theora this = calloc (1, sizeof (*this));
	if (this != NULL){
		this->video_index = -1;
		this->audio_index = -1;
		this->frequency = 44100;
		this->channels = 2;
		this->start = 0;
		this->end = -1;
		this->scaler = 4;
		this->every = 1;
		this->output_width=0;
		this->output_height=0;
		this->video_quality=5;
		this->audio_quality=0;
	}
	return this;
}

void ff2theora_output(ff2theora this) {
	int i;
	AVCodecContext *aenc = NULL;
	AVCodecContext *venc = NULL;
	AVStream *astream = NULL;
	AVStream *vstream = NULL;
	AVCodec *acodec = NULL;
	AVCodec *vcodec = NULL;
	ReSampleContext *resample = NULL;


	double fps = 0.0;

	for (i = 0; i < this->context->nb_streams; i++){
		AVCodecContext *enc = &this->context->streams[i]->codec;
		switch (enc->codec_type){
		case CODEC_TYPE_VIDEO:
			if (this->video_index < 0)
				this->video_index = i;
			break;
		case CODEC_TYPE_AUDIO:
			if (this->audio_index < 0 && !this->disable_audio)
				this->audio_index = i;
			break;
		default:
			break;
		}
	}

	if (this->video_index >= 0){
		vstream = this->context->streams[this->video_index];
		venc = &this->context->streams[this->video_index]->codec;
		vcodec = avcodec_find_decoder (venc->codec_id);
		fps = (double) venc->frame_rate / venc->frame_rate_base;
		if (fps > 10000)
			fps /= 1000;
		if (vcodec == NULL || avcodec_open (venc, vcodec) < 0)
			this->video_index = -1;
		this->fps = fps;

		if(this->output_height>0){
			// we might need that for values other than /16?
			int frame_topBand=0;
			int frame_bottomBand=0;
			int frame_leftBand=0;
			int frame_rightBand=0;
			int frame_padtop=0, frame_padbottom=0;
			int frame_padleft=0, frame_padright=0;
			/* ffmpeg cvs version */
			this->img_resample_ctx = img_resample_full_init(
                                      this->output_width, this->output_height,
                                      venc->width, venc->height,
                                      frame_topBand, frame_bottomBand,
                                      frame_leftBand, frame_rightBand,
				      frame_padtop, frame_padbottom,
				      frame_padleft, frame_padright
				      );
			/* ffmpeg 0.48 */
			/*
			this->img_resample_ctx = img_resample_full_init(
                                      this->output_width, this->output_height,
                                      venc->width, venc->height,
                                      frame_topBand, frame_bottomBand,
                                      frame_leftBand, frame_rightBand
				      );
			*/
			fprintf(stderr,"  Resize: %dx%d => %dx%d\n",venc->width,venc->height,this->output_width,this->output_height);
		}
		else {
			this->output_height=venc->height;
			this->output_width=venc->width;
		}

	}

	if (this->audio_index >= 0){
		astream = this->context->streams[this->audio_index];
		aenc = &this->context->streams[this->audio_index]->codec;
		acodec = avcodec_find_decoder (aenc->codec_id);
		if (this->channels != aenc->channels && aenc->codec_id == CODEC_ID_AC3)
			aenc->channels = this->channels;
		if (acodec != NULL && avcodec_open (aenc, acodec) >= 0)
			resample = audio_resample_init (this->channels,aenc->channels,this->frequency,aenc->sample_rate);
		else
			this->audio_index = -1;
	}

	if (this->video_index >= 0 || this->audio_index >=0){
		AVFrame *frame=NULL;
		AVFrame *frame_tmp=NULL;
		AVFrame *output=NULL;
		AVFrame *output_tmp=NULL;
		AVFrame *output_resized=NULL;
		
		AVPacket pkt;
		int len;
		int len1;
		int got_picture;
		int first = 1;
		uint8_t *ptr;
		int16_t *audio_buf= av_malloc(4*AVCODEC_MAX_AUDIO_FRAME_SIZE);
		int16_t *resampled= av_malloc(4*AVCODEC_MAX_AUDIO_FRAME_SIZE);
		
		
		if(this->video_index >= 0)
			info.audio_only=0;
		else
			info.audio_only=1;
		
		int ret;
		if(!info.audio_only){
			frame = alloc_picture(vstream->codec.pix_fmt,
							vstream->codec.width,vstream->codec.height);
			frame_tmp = alloc_picture(vstream->codec.pix_fmt,
							vstream->codec.width,vstream->codec.height);
			output_tmp =alloc_picture(PIX_FMT_YUV420P, 
							vstream->codec.width,vstream->codec.height);
			output =alloc_picture(PIX_FMT_YUV420P, 
							vstream->codec.width,vstream->codec.height);
			output_resized =alloc_picture(PIX_FMT_YUV420P, 
							this->output_width,this->output_height);
		}

		if(!info.audio_only){
			/* video settings here */
			/*  config file? commandline options? v2v presets? */
			theora_info_init (&info.ti);
			info.ti.width = this->output_width;
			info.ti.height = this->output_height;
			info.ti.frame_width = this->output_width;
			info.ti.frame_height = this->output_height;
			info.ti.offset_x = 0;
			info.ti.offset_y = 0;
			info.ti.fps_numerator = 1000000 * (this->fps);	/* fps= numerator/denominator */
			info.ti.fps_denominator = 1000000;
			info.ti.aspect_numerator=4;
			info.ti.aspect_denominator=3;
			info.ti.colorspace = OC_CS_UNSPECIFIED;
			//info.ti.target_bitrate=1200; 
			info.ti.quality = this->video_quality;
			info.ti.dropframes_p = 0;
			info.ti.quick_p = 1;
			info.ti.keyframe_auto_p = 1;
			info.ti.keyframe_frequency = 64;
			info.ti.keyframe_frequency_force = 64;
			info.ti.keyframe_data_target_bitrate = info.ti.target_bitrate * 1.5;
			info.ti.keyframe_auto_threshold = 80;
			info.ti.keyframe_mindistance = 8;
			info.ti.noise_sensitivity = 1;
		}
		/* audio settings here */
		info.channels = this->channels;
		info.frequency = this->frequency;
		info.vorbis_quality = this->audio_quality;
		theoraframes_init ();
		
		do{
			ret = av_read_packet (this->context, &pkt);
			
			ptr = pkt.data;
			len = pkt.size;

			if (ret >= 0 && pkt.stream_index == this->video_index){
	
				if(len == 0 && !first){
					fprintf (stderr, "no frame available\n");
				}
				while(len > 0){
					len1 = avcodec_decode_video(&vstream->codec, frame,&got_picture, ptr, len);
					if(len1 < 0)
						break;

					if(got_picture){
						/* might have to cast other progressive formats here */
						if(venc->pix_fmt != PIX_FMT_YUV420P){
							img_convert((AVPicture *)output,PIX_FMT_YUV420P,
										(AVPicture *)frame,venc->pix_fmt,
										venc->width,venc->height);
							
							if(avpicture_deinterlace((AVPicture *)output_tmp,
											(AVPicture *)output,PIX_FMT_YUV420P
											,venc->width,venc->height)<0){
								output_tmp = output;
							}
						}
						else{
							/* there must be better way to do this, it seems to work like this though */
							if(frame->linesize[0] != vstream->codec.width) {
								img_convert((AVPicture *)output_tmp,PIX_FMT_YUV420P,
										(AVPicture *)frame,venc->pix_fmt,venc->width,venc->height);
							}
							else {
								output_tmp=frame;
							}
						}
						// now output_tmp
						if(this->img_resample_ctx){
							img_resample(this->img_resample_ctx, 
										(AVPicture *)output_resized, (AVPicture *)output_tmp);
						}	
						else{
							output_resized=output_tmp;
						}
						// now output_resized
						first = 0;
						if(theoraframes_add_video(output_resized->data[0],
							this->output_width,this->output_height,output_resized->linesize[0])){
							ret = -1;
							fprintf (stderr,"No theora frames available\n");
							break;
						}
					}
					ptr += len1;
					len -= len1;
				}
			}
			else if(ret >= 0 && pkt.stream_index == this->audio_index){
				int data_size;
				while(len > 0 ){
					len1 = avcodec_decode_audio(&astream->codec, audio_buf,&data_size, ptr, len);
					if (len1 < 0 || data_size ==0) {
                		/* if error, we skip the frame */
                		break;
            		}	

					len -= len1;
					ptr += len1;
					if(data_size >0){
						int samples =data_size / (aenc->channels * 2);
						int samples_out = audio_resample(resample, resampled, audio_buf, samples);
						if (theoraframes_add_audio(resampled, samples_out *(aenc->channels),samples_out)){
							ret = -1;
							fprintf (stderr,"No audio frames available\n");
							break;
						}
					}
				}
			}
			
			/* flush out the file */
			theoraframes_flush ();
			av_free_packet (&pkt);
		}
		while (ret >= 0);

		av_free(audio_buf);

		if (this->img_resample_ctx)
		    img_resample_close(this->img_resample_ctx);
		if (this->audio_resample)
		    audio_resample_close(resample);

		av_free(resampled);
		theoraframes_close ();
	}
	else{
		fprintf (stderr, "No video or audio stream found\n");
	}
}

void ff2theora_close (ff2theora this){
	av_free (this);
	/* clear out state */
}

void show_stats (AVFormatContext * ic, const char *url, int fps){
	int i;

	printf ("T: ");
	if (ic->duration != AV_NOPTS_VALUE){
		int hours, mins, secs, us;
		secs = ic->duration / AV_TIME_BASE;
		us = ic->duration % AV_TIME_BASE;
		mins = secs / 60;
		secs %= 60;
		hours = mins / 60;
		mins %= 60;
		printf ("%02d:%02d:%02d.%01d\n", hours, mins, secs,(10 * us) / AV_TIME_BASE);
	}
	else{
		printf ("N/A\n");
	}

	printf ("F: ");
	if (ic->duration != AV_NOPTS_VALUE){
		double frames = ((ic->duration / AV_TIME_BASE) * fps);
		printf ("%.0f\n", frames);
	}
	else{
		printf ("N/A\n");
	}

	for (i = 0; i < ic->nb_streams; i++){
		AVStream *st = ic->streams[i];
		AVCodecContext *enc = &st->codec;
		AVCodec *p = avcodec_find_decoder (enc->codec_id);
		if (p)
			printf ("%c%d: %s\n",enc->codec_type ==CODEC_TYPE_VIDEO ? 'V' : 'A', i, p->name);
	}
}

void print_usage (){
	fprintf (stderr, "ffmpeg2theora [options] input\n"
		" Options:\n"
		"\t --output,-o alternative output\n"
		"\t --format,-f define input format\n"
		"\t --width, -x scale to given size\n"
		"\t --height,-y scale to given size\n"
		"\t --videoquality,-v encoding quality for video\n"
		"\t --audioquality,-a encoding quality for audio\n"
		"\t --help,-h this message\n"
		"\n Examples:\n"
		"\tcat something.dv | ffmpeg2theora -f dv -o output.ogg -\n\n"
		"\tLive encoding from a DV camcorder (needs a fast machine)\n"
		"\tdvgrab - | ffmpeg2theora -f dv -x 352 -y 288 -o output.ogg -\n"
		"\n");
	exit (0);
}

int main (int argc, char **argv){
	int stats = 0;
	int outputfile_set=0;
	char outputfile_name[255];
	char inputfile_name[255];
	AVInputFormat *input_fmt=NULL;
	ff2theora convert = ff2theora_init ();
	av_register_all ();
	
     int c,long_option_index;
	const char *optstring = "o:f:x:y:v:a:h::";
	struct option options [] = {
	  {"output",required_argument,NULL,'o'},
	  {"format",required_argument,NULL,'f'},
	  {"width",required_argument,NULL,'x'},
	  {"height",required_argument,NULL,'y'},
	  {"videoquality",required_argument,NULL,'v'},
	  {"audioquality",required_argument,NULL,'a'},
	  {"help",NULL,NULL,'h'},
	  {NULL,0,NULL,0}
	};
	if (argc == 1){
		print_usage ();
	}
	
	while((c=getopt_long(argc,argv,optstring,options,&long_option_index))!=EOF){
		switch(c)
	    {
			case 'o':
				sprintf(outputfile_name,optarg);
				outputfile_set=1;
				break;
			case 'f':
				input_fmt=av_find_input_format(optarg);
				break;
			case 'x':
				convert->output_width=atoi(optarg);
				break;
			case 'y':
				convert->output_height=atoi(optarg);
				break;
			case 'v':
				convert->video_quality=atoi(optarg);
				break;
			case 'a':
				convert->audio_quality=atoi(optarg);
				break;
			case 'h':
				print_usage ();
				exit(1);
		}  
	}	
	
	while(optind<argc){
		/* assume that anything following the options must be a filename */
		if(!strcmp(argv[optind],"-")){
			sprintf(inputfile_name,"pipe:");
		}
		else{
			//fprintf(stderr,"%s.\n",argv[optind]);	
			sprintf(inputfile_name,"%s",argv[optind]);
			if(outputfile_set!=1){
				sprintf(outputfile_name,"%s.ogg",argv[optind]);
				outputfile_set=1;
			}	
		}
		optind++;
	}
	using_stdin |= !strcmp(inputfile_name, "pipe:" ) ||
                   !strcmp( inputfile_name, "/dev/stdin" );

	if(outputfile_set!=1){	
		fprintf(stderr,"output not specified.\n");	
		exit(1);
	}
	//fprintf(stderr,"config done: in='%s', out='%s'.\n",inputfile_name,outputfile_name);
	if (av_open_input_file(&convert->context, inputfile_name, input_fmt, 0, NULL) >= 0){
			outfile = fopen(outputfile_name,"wb");
			if (av_find_stream_info (convert->context) >= 0){
				if (!stats){
					dump_format (convert->context, 0,inputfile_name, 0);
					ff2theora_output (convert);
					convert->audio_index =convert->video_index = -1;
				}
				else{
					show_stats (convert->context, inputfile_name,convert->fps);
				}
			}
			else{
				fprintf (stderr,"Unable to obtain stream info\n");
			}
			av_close_input_file (convert->context);
		}
		else{
			fprintf (stderr, "Unable to open file %s\n", inputfile_name);
		}
	ff2theora_close (convert);
	fprintf(stderr,"\n");
	return(0);
}
